@@ -38,7 +38,7 @@
 #include <math.h>
 #include <float.h>
 
-#define TABLE_SIZE		(32771)		/* Should be a prime */
+#define TABLE_SIZE		(1009)		/* Should be a prime */
 
 #define OPT_QUIET		(0x00000001)
 #define OPT_CUMULATIVE		(0x00000002)
@@ -51,44 +51,69 @@
 #define OPT_KERNEL		(0x00000100)
 #define OPT_USER		(0x00000200)
 #define OPT_CMD			(OPT_CMD_SHORT | OPT_CMD_LONG)
+#define OPT_SHOW_WHENCE		(0x00000400)
 
-typedef struct link {
-	void *data;			/* Data in list */
-	struct link *next;		/* Next item in list */
-} link_t;
+#define EVENT_BUF_SIZE		(8192)
+#define TIMER_REAP_AGE		(600)		/* How old a timer is before we reap it */
+#define TIMER_REAP_THRESHOLD	(30)
 
-typedef struct {
-	link_t	*head;			/* Head of list */
-	link_t	*tail;			/* Tail of list */
-	size_t	length;			/* Length of list */
-} list_t;
+#define _VER_(major, minor, patchlevel) \
+	((major * 10000) + (minor * 100) + patchlevel)
 
-typedef void (*list_link_free_t)(void *);
+#if defined(__GNUC__) && defined(__GNUC_MINOR__)
+#if defined(__GNUC_PATCHLEVEL__)
+#define NEED_GNUC(major, minor, patchlevel) \
+	_VER_(major, minor, patchlevel) <= _VER_(__GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__)
+#else
+#define NEED_GNUC(major, minor, patchlevel) \
+	_VER_(major, minor, patchlevel) <= _VER_(__GNUC__, __GNUC_MINOR__, 0)
+#endif
+#else
+#define NEED_GNUC(major, minor, patchlevel) (0)
+#endif
+
+#if defined(__GNUC__) && NEED_GNUC(4,6,0)
+#define HOT __attribute__ ((hot))
+#else
+#define HOT
+#endif
+
+#if defined(__GNUC__) && !defined(__clang__) && NEED_GNUC(4,6,0)
+#define OPTIMIZE3 __attribute__((optimize("-O3")))
+#else
+#define OPTIMIZE3
+#endif
 
 typedef struct timer_info {
+	struct timer_info *next;	/* Next in list */
+	struct timer_info *hash_next;	/* Next in hash list */
 	pid_t		pid;
 	char 		*task;		/* Name of process/kernel task */
+	char 		*task_mangled;	/* Modified name of process/kernel task */
 	char		*cmdline;	/* From /proc/$pid/cmdline */
 	char		*func;		/* Kernel waiting func */
 	char		*callback;	/* Kernel timer callback func */
 	char		*ident;		/* Unique identity */
 	bool		kernel_thread;	/* True if task is a kernel thread */
+	uint32_t	ref_count;	/* timer stat reference count */
 	uint64_t	total;		/* Total number of events */
 	double		time_total;	/* Total time */
+	double		last_used;	/* Last referenced */
 } timer_info_t;
 
 typedef struct timer_stat {
+	struct timer_stat *next;	/* Next timer stat in hash table */
+	struct timer_stat *sorted_freq_next; /* Next timer stat in event frequency sorted list */
 	uint64_t	count;		/* Number of events */
 	int64_t		delta;		/* Change in events since last time */
 	double		time;		/* Time of sample */
 	double		time_delta;	/* Change in time since last time */
 	timer_info_t	*info;		/* Timer info */
-	struct timer_stat *next;	/* Next timer stat in hash table */
-	struct timer_stat *sorted_freq_next; /* Next timer stat in event frequency sorted list */
 } timer_stat_t;
 
 /* sample delta item as an element of the sample_delta_list_t */
 typedef struct sample_delta_item {
+	struct sample_delta_item *next;	/* next in list */
 	int64_t		delta;		/* difference in timer events between old and new */
 	double		time_delta;	/* difference in time between old and new */
 	timer_info_t	*info;		/* timer this refers to */
@@ -96,8 +121,9 @@ typedef struct sample_delta_item {
 
 /* list of sample_delta_items */
 typedef struct sample_delta_list {
+	struct sample_delta_list *next;	/* next in list */
+	struct sample_delta_item *list;	/* list of sample delta items */
 	double		whence;		/* when the sample was taken */
-	list_t		list;
 } sample_delta_list_t;
 
 typedef struct {
@@ -110,13 +136,17 @@ typedef struct {
 static const char *app_name = "eventstat";
 static const char *proc_timer_stats = "/proc/timer_stats";
 
-static list_t timer_info_list;		/* cache list of timer_info */
-static list_t sample_list;		/* list of samples, sorted in sample time order */
+static timer_stat_t *timer_stat_free_list; /* free list of timer stats */
+static timer_info_t *timer_info_list;	/* cache list of timer_info */
+static uint32_t timer_info_list_length;	/* length of timer_info_list */
+static timer_info_t *timer_info_hash[TABLE_SIZE]; /* hash of timer_info */
+static sample_delta_list_t *sample_delta_list;	/* list of samples, sorted in sample time order */
 static char *csv_results;		/* results in comma separated values */
 static volatile bool stop_eventstat = false;	/* set by sighandler */
 static double  opt_threshold;		/* ignore samples with event delta less than this */
 static uint32_t opt_flags;		/* option flags */
 static bool sane_procs;			/* false if we are in a container */
+static char *get_events_buf;		/* buffer to glob events into */
 
 /*
  *  Attempt to catch a range of signals so
@@ -133,18 +163,9 @@ static const int signals[] = {
 #ifdef SIGQUIT
 	SIGQUIT,
 #endif
-#ifdef SIGILL
-	SIGILL,
-#endif
-#ifdef SIGABRT
-	SIGABRT,
-#endif
 #ifdef SIGFPE
 	SIGFPE,
 #endif
-#ifdef SIGSEGV
-	SIGSEGV,
-#endif
 #ifdef SIGTERM
 	SIGTERM,
 #endif
@@ -155,9 +176,6 @@ static const int signals[] = {
 	SIGUSR2,
 	/* POSIX.1-2001 */
 #endif
-#ifdef SIGBUS
-	SIGBUS,
-#endif
 #ifdef SIGXCPU
 	SIGXCPU,
 #endif
@@ -183,6 +201,22 @@ static const int signals[] = {
 	-1,
 };
 
+/*
+ *  hash_djb2a()
+ *	Hash a string, from Dan Bernstein comp.lang.c (xor version)
+ */
+static HOT OPTIMIZE3 uint32_t hash_djb2a(const char *str)
+{
+	register uint32_t hash = 5381;
+	register int c;
+
+	while ((c = *str++)) {
+		/* (hash * 33) ^ c */
+		hash = ((hash << 5) + hash) ^ c;
+	}
+	return hash % TABLE_SIZE;
+}
+
 /*
  *  set_timer_stat()
  *	enable/disable timer stat
@@ -190,7 +224,7 @@ static const int signals[] = {
 static void set_timer_stat(const char *str, const bool carp)
 {
 	int fd;
-	ssize_t len = (ssize_t)strlen(str);
+	const ssize_t len = (ssize_t)strlen(str);
 
 	if ((fd = open(proc_timer_stats, O_WRONLY, S_IRUSR | S_IWUSR)) < 0) {
 		if (carp) {
@@ -215,19 +249,13 @@ static void set_timer_stat(const char *str, const bool carp)
 }
 
 
-/*
- *  Stop gcc complaining about no return func
- */
-static void eventstat_exit(const int status) __attribute__ ((noreturn));
-
 /*
  *  eventstat_exit()
  *	exit and set timer stat to 0
  */
-static void eventstat_exit(const int status)
+static void __attribute__ ((noreturn)) eventstat_exit(const int status)
 {
 	set_timer_stat("0\n", false);
-
 	exit(status);
 }
 
@@ -241,7 +269,6 @@ static inline double timeval_to_double(const struct timeval *const tv)
 	return (double)tv->tv_sec + ((double)tv->tv_usec / 1000000.0);
 }
 
-
 /*
  *  double_to_timeval
  *	seconds in double to timeval
@@ -252,7 +279,6 @@ static inline struct timeval double_to_timeval(const double val)
 
 	tv.tv_sec = val;
 	tv.tv_usec = (val - (time_t)val) * 1000000.0;
-
 	return tv;
 }
 
@@ -313,61 +339,6 @@ static bool sane_proc_pid_info(void)
 	return ret;
 }
 
-/*
- *  list_init()
- *	initialize list
- */
-static inline void list_init(list_t *list)
-{
-	list->head = NULL;
-	list->tail = NULL;
-	list->length = 0;
-}
-
-/*
- *  list_append()
- *	add a new item to end of the list
- */
-static link_t *list_append(list_t *list, void *data)
-{
-	link_t *link;
-
-	if ((link = calloc(1, sizeof(link_t))) == NULL) {
-		fprintf(stderr, "Cannot allocate list link\n");
-		eventstat_exit(EXIT_FAILURE);
-	}
-	link->data = data;
-
-	if (list->head == NULL) {
-		list->head = link;
-	} else {
-		list->tail->next = link;
-	}
-	list->tail = link;
-	list->length++;
-
-	return link;
-}
-
-/*
- *  list_free()
- *	free the list
- */
-static void list_free(list_t *list, const list_link_free_t freefunc)
-{
-	link_t	*link, *next;
-
-	if (list == NULL)
-		return;
-
-	for (link = list->head; link; link = next) {
-		next = link->next;
-		if (link->data && freefunc)
-			freefunc(link->data);
-		free(link);
-	}
-}
-
 /*
  *  handle_sig()
  *      catch signal, flag a stop and restore timer stat
@@ -380,25 +351,25 @@ static void handle_sig(int dummy)
 	set_timer_stat("0\n", false);
 }
 
-/*
- *  sample_delta_free()
- *	free the sample delta list
- */
-static void sample_delta_free(void *data)
-{
-	sample_delta_list_t *sdl = (sample_delta_list_t*)data;
-
-	list_free(&sdl->list, free);
-	free(sdl);
-}
-
 /*
  *  samples_free()
  *	free collected samples
  */
-static void samples_free(void)
+static inline void samples_free(void)
 {
-	list_free(&sample_list, sample_delta_free);
+	sample_delta_list_t *sdl = sample_delta_list;
+
+	while (sdl) {
+		sample_delta_list_t *sdl_next = sdl->next;
+		sample_delta_item_t *sdi = sdl->list;
+		while (sdi) {
+			sample_delta_item_t *sdi_next = sdi->next;
+			free(sdi);
+			sdi = sdi_next;
+		}
+		free(sdl);
+		sdl = sdl_next;
+	}
 }
 
 /*
@@ -409,16 +380,14 @@ static void sample_add(
 	timer_stat_t *timer_stat,
 	const double whence)
 {
-	link_t	*link;
 	bool	found = false;
-	sample_delta_list_t *sdl = NULL;
+	sample_delta_list_t *sdl;
 	sample_delta_item_t *sdi;
 
 	if (csv_results == NULL)	/* No need if not request */
 		return;
 
-	for (link = sample_list.head; link; link = link->next) {
-		sdl = (sample_delta_list_t*)link->data;
+	for (sdl = sample_delta_list; sdl; sdl = sdl->next) {
 		if (sdl->whence == whence) {
 			found = true;
 			break;
@@ -435,7 +404,8 @@ static void sample_add(
 			eventstat_exit(EXIT_FAILURE);
 		}
 		sdl->whence = whence;
-		list_append(&sample_list, sdl);
+		sdl->next = sample_delta_list;
+		sample_delta_list = sdl;
 	}
 
 	/* Now append the sdi onto the list */
@@ -447,19 +417,19 @@ static void sample_add(
 	sdi->time_delta = timer_stat->time_delta;
 	sdi->info  = timer_stat->info;
 
-	list_append(&sdl->list, sdi);
+	sdi->next = sdl->list;
+	sdl->list = sdi;
 }
 
 /*
  *  sample_find()
  *	scan through a sample_delta_list for timer info, return NULL if not found
  */
-inline static sample_delta_item_t *sample_find(sample_delta_list_t *sdl, const timer_info_t *info)
+inline HOT static sample_delta_item_t *sample_find(sample_delta_list_t *sdl, const timer_info_t *info)
 {
-	link_t *link;
+	sample_delta_item_t *sdi;
 
-	for (link = sdl->list.head; link; link = link->next) {
-		sample_delta_item_t *sdi = (sample_delta_item_t*)link->data;
+	for (sdi = sdl->list; sdi; sdi = sdi->next) {
 		if (sdi->info == info)
 			return sdi;
 	}
@@ -481,62 +451,66 @@ static int info_compare_total(const void *item1, const void *item2)
 	return ((*info2)->total > (*info1)->total) ? 1 : -1;
 }
 
+static bool pid_a_kernel_thread_guess(const char *task)
+{
+	/*
+	 * This is not exactly accurate, but if we can't look up
+	 * a process then try and infer something from the comm field.
+	 * Until we have better kernel support to map /proc/timer_stats
+	 * pids to containerised pids this is the best we can do.
+	 */
+	static const kernel_task_info kernel_tasks[] = {
+		KERN_TASK_INFO("swapper/"),
+		KERN_TASK_INFO("kworker/"),
+		KERN_TASK_INFO("ksoftirqd/"),
+		KERN_TASK_INFO("watchdog/"),
+		KERN_TASK_INFO("migration/"),
+		KERN_TASK_INFO("irq/"),
+		KERN_TASK_INFO("mmcqd/"),
+		KERN_TASK_INFO("jbd2/"),
+		KERN_TASK_INFO("kthreadd"),
+		KERN_TASK_INFO("kthrotld"),
+		KERN_TASK_INFO("kswapd"),
+		KERN_TASK_INFO("ecryptfs-kthrea"),
+		KERN_TASK_INFO("kauditd"),
+		KERN_TASK_INFO("kblockd"),
+		KERN_TASK_INFO("kcryptd"),
+		KERN_TASK_INFO("kdevtmpfs"),
+		KERN_TASK_INFO("khelper"),
+		KERN_TASK_INFO("khubd"),
+		KERN_TASK_INFO("khugepaged"),
+		KERN_TASK_INFO("khungtaskd"),
+		KERN_TASK_INFO("flush-"),
+		KERN_TASK_INFO("bdi-default-"),
+		{ NULL, 0 }
+	};
+	size_t i;
+
+	for (i = 0; kernel_tasks[i].task != NULL; i++) {
+		if (strncmp(task, kernel_tasks[i].task, kernel_tasks[i].len) == 0)
+			return true;
+	}
+	return false;
+}
+
 /*
  *  pid_a_kernel_thread
  *
  */
-static bool pid_a_kernel_thread(const char *task, const pid_t id)
+static bool pid_a_kernel_thread(const char *cmdline, const char *task, const pid_t id)
 {
-	if (sane_procs) {
-		return getpgid(id) == 0;
-	} else {
-		/* In side a container, make a guess at kernel threads */
-		int i;
-		pid_t pgid = getpgid(id);
+	pid_t pgid;
 
-		/* This fails for kernel threads inside a container */
-		if (pgid >= 0)
-			return pgid == 0;
+	if (sane_procs && (cmdline != NULL))
+		return (*cmdline == '\0');
 
-		/*
-		 * This is not exactly accurate, but if we can't look up
-		 * a process then try and infer something from the comm field.
-		 * Until we have better kernel support to map /proc/timer_stats
-		 * pids to containerised pids this is the best we can do.
-		 */
-		static kernel_task_info kernel_tasks[] = {
-			KERN_TASK_INFO("swapper/"),
-			KERN_TASK_INFO("kworker/"),
-			KERN_TASK_INFO("ksoftirqd/"),
-			KERN_TASK_INFO("watchdog/"),
-			KERN_TASK_INFO("migration/"),
-			KERN_TASK_INFO("irq/"),
-			KERN_TASK_INFO("mmcqd/"),
-			KERN_TASK_INFO("jbd2/"),
-			KERN_TASK_INFO("kthreadd"),
-			KERN_TASK_INFO("kthrotld"),
-			KERN_TASK_INFO("kswapd"),
-			KERN_TASK_INFO("ecryptfs-kthrea"),
-			KERN_TASK_INFO("kauditd"),
-			KERN_TASK_INFO("kblockd"),
-			KERN_TASK_INFO("kcryptd"),
-			KERN_TASK_INFO("kdevtmpfs"),
-			KERN_TASK_INFO("khelper"),
-			KERN_TASK_INFO("khubd"),
-			KERN_TASK_INFO("khugepaged"),
-			KERN_TASK_INFO("khungtaskd"),
-			KERN_TASK_INFO("flush-"),
-			KERN_TASK_INFO("bdi-default-"),
-			{ NULL, 0 }
-		};
+	/* We are either in a container, or with a task with a NULL cmdline */
+	pgid = getpgid(id);
+	if (pgid >= 0)
+		return (pgid == 0);
 
-		for (i = 0; kernel_tasks[i].task != NULL; i++) {
-			if (strncmp(task, kernel_tasks[i].task, kernel_tasks[i].len) == 0)
-				return true;
-		}
-	}
-
-	return false;
+	/* Can't get pgid on that pid, so make a guess */
+	return pid_a_kernel_thread_guess(task);
 }
 
 /*
@@ -555,11 +529,12 @@ static char *get_pid_cmdline(const pid_t id)
 	if ((fd = open(buffer, O_RDONLY)) < 0)
 		return NULL;
 
-	if ((ret = read(fd, buffer, sizeof(buffer))) <= 0) {
-		(void)close(fd);
-		return NULL;
-	}
+	ret = read(fd, buffer, sizeof(buffer));
 	(void)close(fd);
+	if (ret < 0)
+		return NULL;
+	if (ret == 0)
+		return strdup("");
 
 	buffer[sizeof(buffer)-1] = '\0';
 
@@ -600,15 +575,15 @@ static inline double duration_round(const double duration)
  */
 static void samples_dump(const char *filename)
 {
-	sample_delta_list_t	*sdl;
 	timer_info_t **sorted_timer_infos;
-	link_t	*link;
 	size_t i = 0;
-	size_t n = timer_info_list.length;
+	size_t n;
 	FILE *fp;
 	uint64_t count = 0;
 	double first_time = -1.0;
 	double duration;
+	timer_info_t *info;
+	sample_delta_list_t *sdl;
 
 	if (filename == NULL)
 		return;
@@ -618,60 +593,63 @@ static void samples_dump(const char *filename)
 		return;
 	}
 
-	if ((sorted_timer_infos = calloc(n, sizeof(timer_info_t*))) == NULL) {
+	if ((sorted_timer_infos = calloc(timer_info_list_length, sizeof(timer_info_t*))) == NULL) {
 		fprintf(stderr, "Cannot allocate buffer for sorting timer_infos\n");
 		eventstat_exit(EXIT_FAILURE);
 	}
 
 	/* Just want the timers with some non-zero total */
-	for (n = 0, link = timer_info_list.head; link; link = link->next) {
-		timer_info_t *info = (timer_info_t*)link->data;
+	for (n = 0, info = timer_info_list; info; info = info->next) {
 		if (info->total > 0)
 			sorted_timer_infos[n++] = info;
 	}
 
 	qsort(sorted_timer_infos, n, sizeof(timer_info_t *), info_compare_total);
 
-	fprintf(fp, "Task:");
+	fprintf(fp, "Time:,Task:");
 	for (i = 0; i < n; i++) {
 		char *task;
 
-		if ((opt_flags & OPT_CMD) && (sorted_timer_infos[i]->cmdline != NULL))
+		if (opt_flags & OPT_CMD)
 			task = sorted_timer_infos[i]->cmdline;
 		else
-			task = sorted_timer_infos[i]->task;
+			task = sorted_timer_infos[i]->task_mangled;
 
 		fprintf(fp, ",%s", task);
 	}
 	fprintf(fp, "\n");
 
-	fprintf(fp, "Init Function:");
+	fprintf(fp, ",Init Function:");
 	for (i = 0; i < n; i++)
 		fprintf(fp, ",%s", sorted_timer_infos[i]->func);
 	fprintf(fp, "\n");
 
-	fprintf(fp, "Callback:");
+	fprintf(fp, ",Callback:");
 	for (i = 0; i < n; i++)
 		fprintf(fp, ",%s", sorted_timer_infos[i]->callback);
 	fprintf(fp, "\n");
 
-	fprintf(fp, "Total:");
+	fprintf(fp, ",Total:");
 	for (i = 0; i < n; i++)
 		fprintf(fp, ",%" PRIu64, sorted_timer_infos[i]->total);
 	fprintf(fp, "\n");
 
-	for (link = sample_list.head; link; link = link->next) {
+	for (sdl = sample_delta_list; sdl; sdl= sdl->next) {
+		time_t t = (time_t)sdl->whence;
+		struct tm *tm;
+
 		count++;
-		sdl = (sample_delta_list_t*)link->data;
+		tm = localtime(&t);
+		fprintf(fp, "%2.2d:%2.2d:%2.2d",
+			tm->tm_hour, tm->tm_min, tm->tm_sec);
 
 		if (first_time < 0)
 			first_time = sdl->whence;
-		fprintf(fp, "%f", duration_round(sdl->whence - first_time));
+		fprintf(fp, ",%f", duration_round(sdl->whence - first_time));
 
 		/* Scan in timer info order to be consistent for all sdl rows */
 		for (i = 0; i < n; i++) {
 			sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
-
 			/*
 			 *  duration - if -C option is used then don't scale by the
 			 *  per sample duration time, instead give the raw sample count
@@ -691,12 +669,12 @@ static void samples_dump(const char *filename)
 	 *  -S option - some statistics, min, max, average, std.dev.
 	 */
 	if (opt_flags & OPT_RESULT_STATS) {
-		fprintf(fp, "Min:");
+		fprintf(fp, ",Min:");
 		for (i = 0; i < n; i++) {
 			double min = DBL_MAX;
+			sample_delta_list_t *sdl;
 
-			for (link = sample_list.head; link; link = link->next) {
-				sdl = (sample_delta_list_t*)link->data;
+			for (sdl = sample_delta_list; sdl; sdl = sdl->next) {
 				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
 
 				if (sdi) {
@@ -710,12 +688,12 @@ static void samples_dump(const char *filename)
 		}
 		fprintf(fp, "\n");
 
-		fprintf(fp, "Max:");
+		fprintf(fp, ",Max:");
 		for (i = 0; i < n; i++) {
 			double max = DBL_MIN;
+			sample_delta_list_t *sdl;
 
-			for (link = sample_list.head; link; link = link->next) {
-				sdl = (sample_delta_list_t*)link->data;
+			for (sdl = sample_delta_list; sdl; sdl= sdl->next) {
 				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
 
 				if (sdi) {
@@ -729,7 +707,7 @@ static void samples_dump(const char *filename)
 		}
 		fprintf(fp, "\n");
 
-		fprintf(fp, "Average:");
+		fprintf(fp, ",Average:");
 		for (i = 0; i < n; i++)
 			fprintf(fp, ",%f", count == 0 ? 0.0 :
 				(double)sorted_timer_infos[i]->total / count);
@@ -738,13 +716,13 @@ static void samples_dump(const char *filename)
 		/*
 		 *  population standard deviation
 		 */
-		fprintf(fp, "Std.Dev.:");
+		fprintf(fp, ",Std.Dev.:");
 		for (i = 0; i < n; i++) {
 			double average = (double)sorted_timer_infos[i]->total / (double)count;
 			double sum = 0.0;
+			sample_delta_list_t *sdl;
 
-			for (link = sample_list.head; link; link = link->next) {
-				sdl = (sample_delta_list_t*)link->data;
+			for (sdl = sample_delta_list; sdl; sdl = sdl->next) {
 				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
 				if (sdi) {
 					double duration = duration_round((opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : sdi->time_delta);
@@ -769,17 +747,20 @@ static void samples_dump(const char *filename)
  *	try to find existing timer info in cache, and to the cache
  *	if it is new.
  */
-static timer_info_t *timer_info_find(const timer_info_t *new_info)
+static HOT timer_info_t *timer_info_find(
+	const timer_info_t *new_info,
+	const char *ident,
+	const double time_now)
 {
-	link_t *link;
 	timer_info_t *info;
+	const uint32_t h = hash_djb2a(ident);
 
-	for (link = timer_info_list.head; link; link = link->next) {
-		info = (timer_info_t*)link->data;
-		if (strcmp(new_info->ident, info->ident) == 0)
+	for (info = timer_info_hash[h]; info; info = info->hash_next) {
+		if (strcmp(ident, info->ident) == 0) {
+			info->last_used = time_now;
 			return info;
+		}
 	}
-
 	if ((info = calloc(1, sizeof(timer_info_t))) == NULL) {
 		fprintf(stderr, "Cannot allocate timer info\n");
 		eventstat_exit(EXIT_FAILURE);
@@ -787,17 +768,20 @@ static timer_info_t *timer_info_find(const timer_info_t *new_info)
 
 	info->pid = new_info->pid;
 	info->task = strdup(new_info->task);
-	if (opt_flags & OPT_CMD)
-		info->cmdline = get_pid_cmdline(new_info->pid);
-
+	info->task_mangled = strdup(new_info->task_mangled);
+	info->cmdline = strdup(new_info->cmdline);
 	info->func = strdup(new_info->func);
 	info->callback = strdup(new_info->callback);
-	info->ident = strdup(new_info->ident);
+	info->ident = strdup(ident);
 	info->kernel_thread = new_info->kernel_thread;
 	info->total = new_info->total;
 	info->time_total = new_info->time_total;
+	info->last_used = time_now;
+	info->ref_count = 0;
 
 	if (info->task == NULL ||
+	    info->task_mangled == NULL ||
+	    info->cmdline == NULL ||
 	    info->func == NULL ||
 	    info->callback == NULL ||
 	    info->ident == NULL) {
@@ -807,7 +791,12 @@ static timer_info_t *timer_info_find(const timer_info_t *new_info)
 
 	/* Does not exist in list, append it */
 
-	list_append(&timer_info_list, info);
+	info->next = timer_info_list;
+	timer_info_list = info;
+	timer_info_list_length++;
+
+	info->hash_next = timer_info_hash[h];
+	timer_info_hash[h] = info;
 
 	return info;
 }
@@ -821,7 +810,9 @@ static void timer_info_free(void *data)
 	timer_info_t *info = (timer_info_t*)data;
 
 	free(info->task);
-	free(info->cmdline);
+	if (info->cmdline != info->task_mangled)
+		free(info->cmdline);
+	free(info->task_mangled);
 	free(info->func);
 	free(info->callback);
 	free(info->ident);
@@ -829,33 +820,133 @@ static void timer_info_free(void *data)
 }
 
 /*
- *  timer_info_free
- *	free up all unique timer infos
+ *  timer_info_purge_old_from_timer_list()
+ *	remove old timer info from the timer list
  */
-static void timer_info_list_free(void)
+static void timer_info_purge_old_from_timer_list(
+	timer_info_t **list,
+	const double time_now)
 {
-	list_free(&timer_info_list, timer_info_free);
+	timer_info_t *prev = NULL, *info = *list;
+
+	while (info) {
+		timer_info_t *next = info->next;
+
+		/*
+		 * Only remove from list once all timer
+		 * stats no longer reference it
+		 */
+		if ((info->ref_count == 0) &&
+		    (info->last_used + TIMER_REAP_AGE < time_now)) {
+			if (prev == NULL)
+				*list = next;
+			else
+				prev->next = next;
+			timer_info_list_length--;
+		} else {
+			prev = info;
+		}
+		info = next;
+	}
 }
 
 /*
- *  hash_pjw()
- *	Hash a string, from Aho, Sethi, Ullman, Compiling Techniques.
+ *  timer_info_purge_old_from_hash_list()
+ *	remove old timer info from a hash list
  */
-static uint32_t hash_pjw(const char *str)
+static void timer_info_purge_old_from_hash_list(
+	timer_info_t **list,
+	const double time_now)
 {
-  	uint32_t h = 0;
+	timer_info_t *prev = NULL, *info = *list;
 
-	while (*str) {
-		uint32_t g;
-		h = (h << 4) + (*str);
-		if (0 != (g = h & 0xf0000000)) {
-			h = h ^ (g >> 24);
-			h = h ^ g;
+	while (info) {
+		timer_info_t *next = info->hash_next;
+
+		/*
+		 * Only remove and free once all timer stats no
+		 * longer reference it
+		 */
+		if ((info->ref_count == 0) &&
+		    (info->last_used + TIMER_REAP_AGE < time_now)) {
+			if (prev == NULL)
+				*list = next;
+			else
+				prev->hash_next = next;
+
+			timer_info_free(info);
+		} else {
+			prev = info;
 		}
-		str++;
+		info = next;
 	}
+}
 
-  	return h % TABLE_SIZE;
+
+/*
+ *  timer_info_purge_old()
+ *  	clean out old timer infos
+ */
+static inline void timer_info_purge_old(const double time_now)
+{
+	static uint16_t count = 0;
+
+	count++;
+	if (count > TIMER_REAP_THRESHOLD) {
+		size_t i;
+
+		count = 0;
+		timer_info_purge_old_from_timer_list(&timer_info_list, time_now);
+		for (i = 0; i < TABLE_SIZE; i++)
+			timer_info_purge_old_from_hash_list(&timer_info_hash[i], time_now);
+	}
+}
+
+/*
+ *  timer_info_list_free()
+ *	free up all unique timer infos
+ */
+static inline void timer_info_list_free(void)
+{
+	timer_info_t *info = timer_info_list;
+
+	/* Free list and timers on list */
+	while (info) {
+		timer_info_t *next = info->next;
+
+		timer_info_free(info);
+		info = next;
+	}
+}
+
+/*
+ *  make_hash_ident()
+ */
+static char *make_hash_ident(const timer_info_t *info)
+{
+	static char ident[1024];
+
+	snprintf(ident, sizeof(ident), "%x%s%8.8s%8.8s%s",
+		info->pid, info->task, info->func,
+		info->callback, info->cmdline);
+	return ident;
+}
+
+/*
+ *  timer_stat_free_list_free()
+ *	free up the timer stat free list
+ */
+static void timer_stat_free_list_free(void)
+{
+	timer_stat_t *ts = timer_stat_free_list;
+
+	while (ts) {
+		timer_stat_t *next = ts->next;
+
+		free(ts);
+		ts = next;
+	}
+	timer_stat_free_list = NULL;
 }
 
 /*
@@ -872,7 +963,12 @@ static void timer_stat_free_contents(
 
 		while (ts) {
 			timer_stat_t *next = ts->next;
-			free(ts);
+
+			/* Decrement info ref count */
+			ts->info->ref_count--;
+			/* Add it onto the timer stat free list */
+			ts->next = timer_stat_free_list;
+			timer_stat_free_list = ts;
 
 			ts = next;
 		}
@@ -888,47 +984,34 @@ static void timer_stat_free_contents(
 static void timer_stat_add(
 	timer_stat_t *timer_stats[],	/* timer stat hash table */
 	const double time_now,		/* time sample was taken */
-	const uint64_t count,		/* event count */
-	const pid_t pid,		/* PID of task */
-	char *task,			/* Name of task */
-	char *func,			/* Kernel function */
-	char *callback,			/* Kernel timer callback */
-	const bool kernel_thread)	/* Is a kernel thread */
+	timer_info_t *info)		/* timer info to be added */
 {
-	char buf[4096];
-	timer_stat_t *ts;
-	timer_stat_t *ts_new;
-	timer_info_t info;
-	uint32_t h;
-
-	snprintf(buf, sizeof(buf), "%d:%s:%s:%s", pid, task, func, callback);
-	h = hash_pjw(buf);
-	ts = timer_stats[h];
+	const char *ident = make_hash_ident(info);
+	const uint32_t h = hash_djb2a(ident);
+	timer_stat_t *ts, *ts_new;
 
 	for (ts = timer_stats[h]; ts; ts = ts->next) {
-		if (strcmp(ts->info->ident, buf) == 0) {
-			ts->count += count;
+		if (strcmp(ts->info->ident, ident) == 0) {
+			ts->count += info->total;
 			return;
 		}
 	}
-	/* Not found, it is new! */
-
-	if ((ts_new = malloc(sizeof(timer_stat_t))) == NULL) {
-		fprintf(stderr, "Out of memory allocating a timer stat\n");
-		eventstat_exit(EXIT_FAILURE);
+	/* Not found, it is new */
+	if (timer_stat_free_list) {
+		/* Get new one from free list */
+		ts_new = timer_stat_free_list;
+		timer_stat_free_list = timer_stat_free_list->next;
+	} else {
+		/* Get one from heap */
+		if ((ts_new = malloc(sizeof(timer_stat_t))) == NULL) {
+			fprintf(stderr, "Out of memory allocating a timer stat\n");
+			eventstat_exit(EXIT_FAILURE);
+		}
 	}
 
-	info.pid = pid;
-	info.task = task;
-	info.func = func;
-	info.callback = callback;
-	info.ident = buf;
-	info.kernel_thread = kernel_thread;
-	info.total = count;
-	info.time_total = 0.0;
-
-	ts_new->count = count;
-	ts_new->info = timer_info_find(&info);
+	ts_new->count = info->total;
+	ts_new->info = timer_info_find(info, ident, time_now);
+	ts_new->info->ref_count++;
 	ts_new->next = timer_stats[h];
 	ts_new->time = time_now;
 	ts_new->sorted_freq_next = NULL;
@@ -940,19 +1023,15 @@ static void timer_stat_add(
  *  timer_stat_find()
  *	find a timer stat (needle) in a timer stat hash table (haystack)
  */
-static timer_stat_t *timer_stat_find(
+static OPTIMIZE3 timer_stat_t *timer_stat_find(
 	timer_stat_t *haystack[],	/* timer stat hash table */
 	timer_stat_t *needle)		/* timer stat to find */
 {
 	timer_stat_t *ts;
-	char buf[4096];
+	const char *ident = make_hash_ident(needle->info);
 
-	snprintf(buf, sizeof(buf), "%d:%s:%s:%s",
-		needle->info->pid, needle->info->task,
-		needle->info->func, needle->info->callback);
-
-	for (ts = haystack[hash_pjw(buf)]; ts; ts = ts->next) {
-		if (strcmp(ts->info->ident, buf) == 0)
+	for (ts = haystack[hash_djb2a(ident)]; ts; ts = ts->next) {
+		if (strcmp(ts->info->ident, ident) == 0)
 			return ts;
 	}
 
@@ -990,8 +1069,9 @@ static void timer_stat_sort_freq_add(
  *	stats.  We are interested in just current and new timers, not ones that
  *	silently die
  */
-static void timer_stat_diff(
+static OPTIMIZE3 void timer_stat_diff(
 	const double duration,		/* time between each sample */
+	const double time_delta,	/* how long been running sofar */
 	const int32_t n_lines,		/* number of lines to output */
 	const double whence,		/* nth sample */
 	timer_stat_t *timer_stats_old[],/* old timer stats samples */
@@ -1004,9 +1084,11 @@ static void timer_stat_diff(
 		timer_stat_t *ts;
 
 		for (ts = timer_stats_new[i]; ts; ts = ts->next) {
+			ts->info->last_used = whence;
 			timer_stat_t *found =
 				timer_stat_find(timer_stats_old, ts);
 			if (found) {
+				found->info->last_used = whence;
 				ts->delta = ts->count - found->count;
 				ts->time_delta = ts->time - found->time;
 				if (ts->delta >= opt_threshold) {
@@ -1047,16 +1129,15 @@ static void timer_stat_diff(
 					printf("%8.2f ", (double)sorted->delta / duration);
 
 				if (opt_flags & OPT_BRIEF) {
-					char *cmd = sorted->info->cmdline ?
-						sorted->info->cmdline : sorted->info->task;
+					char *cmd = sorted->info->cmdline;
 
 					printf("%5d %s\n",
 						sorted->info->pid,
 						(opt_flags & OPT_CMD) ?
-							cmd : sorted->info->task);
+							cmd : sorted->info->task_mangled);
 				} else {
 					printf("%5d %-15s %-25s %-s\n",
-						sorted->info->pid, sorted->info->task,
+						sorted->info->pid, sorted->info->task_mangled,
 						sorted->info->func, sorted->info->callback);
 				}
 			}
@@ -1071,12 +1152,85 @@ static void timer_stat_diff(
 			total, (double)total / duration,
 			(double)kt_total / duration,
 			(double)(total - kt_total) / duration);
+		if (opt_flags & OPT_SHOW_WHENCE) {
+			time_t t = (time_t)whence;
+			char *timestr = ctime(&t);
+			char *pos = strchr(timestr, '\n');
+
+			if (*pos)
+				*pos = '\0';
+			printf("Timestamp: %s, Total Run Duration: %.1f secs\n", timestr, time_delta);
+		}
+
 		if (!sane_procs)
 			printf("Note: this was run inside a container, kernel tasks were guessed.\n");
 		printf("\n");
 	}
 }
 
+/*
+ *  read_events()
+ *	read in events data into a global read buffer.
+ *	the buffer is auto-expanded where necessary and
+ *	only free'd at exit time.  This way we can parse
+ *	the data a little faster.
+ */
+static char *read_events(void)
+{
+	int fd;
+	static size_t get_events_size;
+	size_t size;
+
+	if (get_events_buf == NULL) {
+		if ((get_events_buf = malloc(EVENT_BUF_SIZE << 1)) == NULL) {
+			fprintf(stderr, "Cannot read %s, out of memory\n", proc_timer_stats);
+			return NULL;
+		}
+		get_events_size = (EVENT_BUF_SIZE << 1);
+	}
+
+	if ((fd = open(proc_timer_stats, O_RDONLY)) < 0) {
+		fprintf(stderr, "Cannot open %s\n", proc_timer_stats);
+		return NULL;
+	}
+
+	size = 0;
+	for (;;) {
+		ssize_t ret;
+		char buffer[EVENT_BUF_SIZE];
+
+		ret = read(fd, buffer, sizeof(buffer));
+		if (ret == 0)
+			break;
+		if (ret < 0) {
+			if ((errno == EINTR) ||
+			    (errno != EAGAIN)) {
+				continue;
+			}
+			break;
+		}
+		/* Do we need to expand the global buffer? */
+		if (size + ret >= get_events_size) {
+			char *tmpptr;
+
+			get_events_size += (EVENT_BUF_SIZE << 1);
+			tmpptr = realloc(get_events_buf, get_events_size + 1);
+			if (!tmpptr) {
+				(void)close(fd);
+				fprintf(stderr, "Cannot read %s, out of memory\n", proc_timer_stats);
+				return NULL;
+			}
+			get_events_buf = tmpptr;
+		}
+		memcpy(get_events_buf + size, buffer, ret);
+		size += ret;
+		*(get_events_buf + size) = '\0';
+	}
+	(void)close(fd);
+
+	return get_events_buf;
+}
+
 /*
  *  get_events()
  *	scan /proc/timer_stats and populate a timer stat hash table with
@@ -1086,70 +1240,77 @@ static void get_events(
 	timer_stat_t *timer_stats[],
 	const double time_now)
 {
-	FILE *fp;
-	char buf[4096];
+	char *tmpptr;
+	const size_t app_name_len = strlen(app_name);
 
-	if ((fp = fopen(proc_timer_stats, "r")) == NULL) {
-		fprintf(stderr, "Cannot open %s\n", proc_timer_stats);
+	if ((tmpptr = read_events()) == NULL)
 		return;
-	}
 
-	/* Originally from PowerTop, but majorly re-worked */
-	while (!feof(fp)) {
-		char *ptr = buf;
-		uint64_t count = 0;
-		pid_t pid = -1;
+	while (*tmpptr) {
+		char *ptr = tmpptr, *eol = tmpptr;
 		char task[64];
-		char func[128];
-		char timer[128];
-		bool kernel_thread;
+		char task_mangled[64];
+		char func[64];
+		char callback[64];
+		char *cmdline;
 		int mask;
+		timer_info_t info;
 
-		if (fgets(buf, sizeof(buf), fp) == NULL)
+		/* Find the end of a line */
+		while (*eol) {
+			if (*eol == '\n') {
+				*eol = '\0';
+				eol++;
+				break;
+			}
+			eol++;
+		}
+
+		/* Is this the last line we want to parse? */
+		if (strstr(tmpptr, "total events") != NULL)
 			break;
 
-		if (strstr(buf, "total events") != NULL)
-			break;
-
-		if (strstr(buf, ",") == NULL)
-			continue;
-
-		/* format: count[D], pid, task, func (timer) */
-
+		/* Looking for a format: "count[D], pid, task, func (timer)" */
 		while (*ptr && *ptr != ',')
 			ptr++;
-
 		if (*ptr != ',')
-			continue;
+			goto next;
+		if (ptr > tmpptr && *(ptr - 1) == 'D')
+			goto next;	/* Deferred event, skip */
 
-		if (ptr > buf && *(ptr-1) == 'D')
-			continue;	/* Deferred event, skip */
+		/* Now we're ready to fetch info fields */
+		memset(&info, 0, sizeof(info));
 
 		ptr++;
-		if (sscanf(buf, "%21" SCNu64, &count) != 1)
-			continue;
+		if (sscanf(tmpptr, "%21" SCNu64, &info.total) != 1)
+			goto next;
 		memset(task, 0, sizeof(task));
 		memset(func, 0, sizeof(func));
-		memset(timer, 0, sizeof(timer));
-		if (sscanf(ptr, "%10d %63s %127s (%127[^)])", &pid, task, func, timer) != 4)
-			continue;
+		memset(callback, 0, sizeof(callback));
+		info.pid = -1;
+		if (sscanf(ptr, "%10d %63s %63s (%63[^)])", &info.pid, task, func, callback) != 4)
+			goto next;
 
-		kernel_thread = pid_a_kernel_thread(task, pid);
+		/* Processes without a command line are kernel threads */
+		cmdline = get_pid_cmdline(info.pid);
+		info.kernel_thread = pid_a_kernel_thread(cmdline, task, info.pid);
 
 		/* Swapper is special, like all corner cases */
 		if (strncmp(task, "swapper", 6) == 0)
-			kernel_thread = true;
-
-		mask = kernel_thread ? OPT_KERNEL : OPT_USER;
+			info.kernel_thread = true;
 
+		mask = info.kernel_thread ? OPT_KERNEL : OPT_USER;
 		if (!(opt_flags & mask))
-			continue;
+			goto free_next;
 
-		if (kernel_thread) {
-			char tmp[64];
-			task[13] = '\0';
-			snprintf(tmp, sizeof(tmp), "[%s]", task);
-			strncpy(task, tmp, 13);
+		if (info.kernel_thread) {
+			char tmp[sizeof(task)];
+
+			strcpy(tmp, task);
+			tmp[13] = '\0';
+			snprintf(task_mangled, sizeof(task_mangled), "[%s]", tmp);
+		} else {
+			strcpy(task_mangled, task);
 		}
 
 		if (strcmp(task, "insmod") == 0)
@@ -1157,15 +1318,23 @@ static void get_events(
 		if (strcmp(task, "modprobe") == 0)
 			strncpy(task, "[kern mod]", 13);
 
-		if ((strncmp(func, "tick_nohz_", 10) == 0) ||
-		    (strncmp(func, "tick_setup_sched_timer", 20) == 0) ||
-		    (strncmp(task, app_name, strlen(app_name)) == 0))
-			continue;
-
-		timer_stat_add(timer_stats, time_now, count, pid, task, func, timer, kernel_thread);
+		if (strncmp(func, "tick_nohz_", 10) &&
+		    strncmp(func, "tick_setup_sched_timer", 20) &&
+		    strncmp(task, app_name, app_name_len)) {
+			info.task = task;
+			info.cmdline = cmdline ? cmdline : task_mangled;
+			info.task_mangled = task_mangled;
+			info.func = func;
+			info.callback = callback;
+			info.ident = tmpptr;
+			info.time_total = 0.0;
+			timer_stat_add(timer_stats, time_now, &info);
+		}
+free_next:
+		free(cmdline);
+next:
+		tmpptr = eol;
 	}
-
-	(void)fclose(fp);
 }
 
 /*
@@ -1188,6 +1357,7 @@ static void show_usage(void)
 	printf("  -s\t\tuse short process name from /proc/pid/cmdline in CSV output.\n");
 	printf("  -S\t\tcalculate min, max, average and standard deviation in CSV output.\n");
 	printf("  -t threshold\tsamples less than the specified threshold are ignored.\n");
+	printf("  -w\t\tadd time stamp (when events occurred) to output.\n");
 }
 
 int main(int argc, char **argv)
@@ -1200,11 +1370,8 @@ int main(int argc, char **argv)
 	struct sigaction new_action;
 	int i;
 
-	list_init(&timer_info_list);
-	list_init(&sample_list);
-
 	for (;;) {
-		int c = getopt(argc, argv, "bcCdksSlhn:qr:t:u");
+		int c = getopt(argc, argv, "bcCdksSlhn:qr:t:uw");
 		if (c == -1)
 			break;
 		switch (c) {
@@ -1263,6 +1430,9 @@ int main(int argc, char **argv)
 		case 'u':
 			opt_flags |= OPT_USER;
 			break;
+		case 'w':
+			opt_flags |= OPT_SHOW_WHENCE;
+			break;
 		default:
 			show_usage();
 			eventstat_exit(EXIT_FAILURE);
@@ -1336,7 +1506,7 @@ int main(int argc, char **argv)
 
 	while (!stop_eventstat && (forever || count--)) {
 		struct timeval tv;
-		double secs, duration = duration_secs;
+		double secs, duration, time_delta;
 		int ret;
 
 		/* Timeout to wait for in the future for this sample */
@@ -1366,15 +1536,18 @@ int main(int argc, char **argv)
 		duration = gettime_to_double() - time_now;
 		duration = floor((duration * 100.0) + 0.5) / 100.0;
 		time_now = gettime_to_double();
+		time_delta = time_now - time_start;
 
 		get_events(timer_stats_new, time_now);
-		timer_stat_diff(duration, n_lines, time_now,
+		timer_stat_diff(duration, time_delta, n_lines, time_now,
 			timer_stats_old, timer_stats_new);
 		timer_stat_free_contents(timer_stats_old);
 
 		tmp             = timer_stats_old;
 		timer_stats_old = timer_stats_new;
 		timer_stats_new = tmp;
+
+		timer_info_purge_old(time_now);
 	}
 abort:
 	samples_dump(csv_results);
@@ -1385,6 +1558,8 @@ abort:
 	free(timer_stats_new);
 	samples_free();
 	timer_info_list_free();
+	timer_stat_free_list_free();
+	free(get_events_buf);
 
 	eventstat_exit(EXIT_SUCCESS);
 }
